package org.psesquared.server.authentication.api.service;

import jakarta.servlet.http.Cookie;
import jakarta.servlet.http.HttpServletResponse;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.NoSuchElementException;
import lombok.RequiredArgsConstructor;
import org.psesquared.server.authentication.api.controller.ChangePasswordRequest;
import org.psesquared.server.authentication.api.controller.PasswordRequest;
import org.psesquared.server.authentication.api.controller.UserInfoRequest;
import org.psesquared.server.authentication.api.data.access.AuthenticationDao;
import org.psesquared.server.config.JwtService;
import org.psesquared.server.model.Role;
import org.psesquared.server.model.User;
import org.springframework.http.HttpStatus;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

/**
 * This service class manages all business logic associated with the
 * authentication API.
 * <br>
 * It is called from the
 * {@link
 * org.psesquared.server.authentication.api.controller.AuthenticationController}
 * and passes on requests concerning data access to the
 * {@link AuthenticationDao}.
 */
@Service
@Transactional
@RequiredArgsConstructor
public class AuthenticationService {

  /**
   * A {@link User} is not enabled until verification.
   */
  private static final boolean ENABLED_DEFAULT = false;

  /**
   * A {@link User} becomes enabled after verification.
   */
  private static final boolean VERIFIED = true;

  /**
   * The age of an expired cookie.
   */
  private static final int EXPIRED_AGE = 0;

  /**
   * The name of the cookie used by podcatchers for authentication.
   * If a {@link User} has logged in, this cookie holds the JWT.
   */
  private static final String COOKIE_NAME = "sessionid";

  /**
   * Specifies that the cookie should be sent to all URLs.
   */
  private static final String COOKIE_PATH_GLOBAL = "/";

  /**
   * The default role of a {@link User} is {@link Role#USER}.
   */
  private static final Role DEFAULT_USER = Role.USER;

  /**
   * The JPA repository that handles all user related database requests.
   */
  private final AuthenticationDao authenticationDao;

  /**
   * The class used for the encryption of passwords.
   */
  private final PasswordEncoder passwordEncoder;

  /**
   * The service class used for managing JWTs.
   */
  private final JwtService jwtService;

  /**
   * The service class used for sending emails.
   */
  private final EmailServiceImpl emailService;

  /**
   * The service class used for encrypting email addresses.
   */
  private final EncryptionService encryptionService;

  /**
   * The service class used for checking if the given user information meets
   * the specified requirements.
   */
  private final InputCheckService inputCheckService;

  /**
   * This method is invoked by the register method of the authentication
   * controller.
   * <br>
   * 1. Checks if the given user information meets the requirements.
   * <br>
   * 2. Checks if no user with the given username already exists (if so,
   * and email as well as password match, the verification email is sent again).
   * <br>
   * 3. Creates a user with the given information and sends verification email.
   *
   * @param userInfo The wrapper object containing username, email and password
   * @return @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#BAD_REQUEST} for invalid user information
   */
  public HttpStatus registerUser(final UserInfoRequest userInfo) {
    if (inputCheckService.checkUsernameInvalid(userInfo.username())
        || inputCheckService.checkEmailInvalid(userInfo.email())
        || inputCheckService.checkPasswordInvalid(userInfo.password())) {
      return HttpStatus.BAD_REQUEST;
    }

    final String encryptedEmailFromRequest
        = encryptionService.saltAndHashEmail(userInfo.email());
    User user;

    try {
      user = authenticationDao
          .findByUsernameOrEmail(userInfo.username(), encryptedEmailFromRequest)
          .orElseThrow();

      if (user.isEnabled()
          || !user.getEmail().equals(encryptedEmailFromRequest)
          || !user.getUsername().equals(userInfo.username())
          || !passwordEncoder
          .matches(userInfo.password(), user.getPassword())) {
        return HttpStatus.BAD_REQUEST;
      }

      user.setCreatedAt(LocalDateTime.now().toEpochSecond(ZoneOffset.UTC));

    } catch (NoSuchElementException e) {
      user = User.builder()
          .username(userInfo.username())
          .email(encryptionService.saltAndHashEmail(userInfo.email()))
          .password(passwordEncoder.encode(userInfo.password()))
          .enabled(ENABLED_DEFAULT)
          .createdAt(LocalDateTime.now().toEpochSecond(ZoneOffset.UTC))
          .role(DEFAULT_USER)
          .build();
      authenticationDao.save(user);
    }
    emailService.sendVerification(userInfo.email(), user);
    return HttpStatus.OK;
  }

  /**
   * This method is invoked by the verifyRegistration method of the
   * authentication controller.
   * <br>
   * If a not yet verified {@link User} with the given username exists,
   * this user is verified via {@link User#setEnabled(boolean)}.
   *
   * @param username The username of the to be verified user
   * @param token The JWT for authentication
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#BAD_REQUEST} user exists and is already verified,
   *        <br>
   *        {@link HttpStatus#UNAUTHORIZED} invalid token,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus verifyRegistration(final String username,
                                       final String token) {
    try {
      var user = authenticationDao.findByUsername(username)
          .orElseThrow();

      if (user.isEnabled()) {
        return HttpStatus.BAD_REQUEST;
      }

      if (!jwtService.isUrlTokenValid(token, user)) {
        return HttpStatus.UNAUTHORIZED;
      }

      user.setEnabled(VERIFIED);
      return HttpStatus.OK;
    } catch (NoSuchElementException e) {
      return HttpStatus.NOT_FOUND;
    }
  }

  /**
   * This method is invoked by the login method of the authentication
   * controller.
   * <br>
   * Sets the "sessionid" cookie with a valid JWT for further authentication.
   *
   * @param username The username of the user who wants to log in
   * @param response The {@link HttpServletResponse} for setting the "sessionid"
   *                 cookie
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus login(final String username,
                          final HttpServletResponse response) {
    try {
      var user = authenticationDao.findByUsername(username)
          .orElseThrow();
      var token = jwtService.generateAccessTokenString(user);
      Cookie cookie = new Cookie(COOKIE_NAME, token);
      cookie.setPath(COOKIE_PATH_GLOBAL);
      response.addCookie(cookie);
      return HttpStatus.OK;
    } catch (NoSuchElementException e) {
      return HttpStatus.NOT_FOUND;
    }
  }

  /**
   * This method is invoked by the logout method of the authentication
   * controller.
   * <br>
   * Invalidates the "sessionid" cookie.
   * Thus, for further authentication until the next login only HTTP basic
   * authentication (and no JWT authentication) is possible.
   *
   * @param username The username of the user who wants to log out
   * @param response The {@link HttpServletResponse} for invalidating the
   *                 "sessionid" cookie
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus logout(final String username,
                           final HttpServletResponse response) {
    if (authenticationDao.existsByUsername(username)) {
      Cookie cookie = new Cookie(COOKIE_NAME, null);
      cookie.setMaxAge(EXPIRED_AGE);
      cookie.setPath(COOKIE_PATH_GLOBAL);
      response.addCookie(cookie);
      return HttpStatus.OK;
    }
    return HttpStatus.NOT_FOUND;
  }

  /**
   * This method is invoked by the forgotPassword method of the authentication
   * controller.
   * <br>
   * Sends an email with a link to reset the password to the given email
   * address, if a user with this email address exists.
   *
   * @param email The email address of the user who wants to reset their
   *              password
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus forgotPassword(final String email) {
    try {
      var user = authenticationDao
          .findByEmail(encryptionService.saltAndHashEmail(email))
          .orElseThrow();
      emailService.sendPasswordReset(email, user);
      return HttpStatus.OK;
    } catch (NoSuchElementException e) {
      return HttpStatus.NOT_FOUND;
    }
  }

  /**
   * This method is invoked by the resetPassword method of the authentication
   * controller.
   * <br>
   * Sets a new password for the given user who has forgotten their
   * password if the JWT is valid.
   *
   * @param username The username of the user who wants to reset their
   *                 password
   * @param token The JWT for authentication
   * @param requestBody The request-wrapper containing the new password
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#BAD_REQUEST} password doesn't meet requirements,
   *        <br>
   *        {@link HttpStatus#UNAUTHORIZED} invalid token,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus resetPassword(final String username,
                                  final String token,
                                  final PasswordRequest requestBody) {
    if (inputCheckService.checkPasswordInvalid(requestBody.password())) {
      return HttpStatus.BAD_REQUEST;
    }
    try {
      var user = authenticationDao.findByUsername(username)
          .orElseThrow();
      if (jwtService.isUrlTokenValid(token, user)) {
        user.setPassword(passwordEncoder.encode(requestBody.password()));
        return HttpStatus.OK;
      }
      return HttpStatus.UNAUTHORIZED;
    } catch (NoSuchElementException e) {
      return HttpStatus.NOT_FOUND;
    }
  }

  /**
   * This method is invoked by the changePassword method of the authentication
   * controller.
   * <br>
   * Changes the password of a logged-in user.
   *
   * @param username The username of the user who wants to change their
   *                 password
   * @param requestBody The request-wrapper containing old and new password
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#BAD_REQUEST} old password is wrong, or new
   *        password doesn't meet requirements,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus changePassword(final String username,
                                   final ChangePasswordRequest requestBody) {
    if (inputCheckService.checkPasswordInvalid(requestBody.newPassword())) {
      return HttpStatus.BAD_REQUEST;
    }
    try {
      var user = authenticationDao.findByUsername(username)
          .orElseThrow();
      if (passwordEncoder
          .matches(requestBody.oldPassword(), user.getPassword())) {
        user.setPassword(passwordEncoder.encode(requestBody.newPassword()));
        return HttpStatus.OK;
      }
      return HttpStatus.BAD_REQUEST;
    } catch (NoSuchElementException e) {
      return HttpStatus.NOT_FOUND;
    }
  }

  /**
   * This method is invoked by the deleteUser method of the authentication
   * controller.
   * <br>
   * Deletes the user with the given username if existent and if the given
   * password for confirmation is correct.
   *
   * @param username The username of the user who wants to delete their account
   * @param requestBody The request-wrapper containing the password for
   *                    confirmation
   * @return {@link HttpStatus#OK} on success,
   *        <br>
   *        {@link HttpStatus#BAD_REQUEST} wrong password,
   *        <br>
   *        {@link HttpStatus#NOT_FOUND} user not found
   */
  public HttpStatus deleteUser(final String username,
                               final PasswordRequest requestBody) {
    try {
      var user = authenticationDao.findByUsername(username)
          .orElseThrow();
      if (passwordEncoder.matches(requestBody.password(), user.getPassword())) {
        authenticationDao.delete(user);
        return HttpStatus.OK;
      }
      return HttpStatus.BAD_REQUEST;
    } catch (NoSuchElementException e) {
      return HttpStatus.NOT_FOUND;
    }
  }

  /**
   * This method is invoked by {@link org.psesquared.server.util.Scheduler}
   * for cleaning the database from expired {@link User}s.
   *
   * @param timestamp The timestamp representing the number of seconds from
   *                  the epoch of 1970-01-01T00:00:00Z.
   * @see AuthenticationDao#deleteAllByEnabledFalseAndCreatedAtLessThan(long)
   */
  public void deleteInvalidUsersOlderThan(final long timestamp) {
    authenticationDao.deleteAllByEnabledFalseAndCreatedAtLessThan(timestamp);
  }

}
