package org.psesquared.server.authentication.api.service;

import io.jsonwebtoken.io.Decoders;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import lombok.RequiredArgsConstructor;
import org.psesquared.server.config.SecurityConfigProperties;
import org.springframework.stereotype.Service;

/**
 * The service class responsible for encrypting the email addresses of
 * {@link org.psesquared.server.model.User}s.
 */
@Service
@RequiredArgsConstructor
public class EncryptionService {

  /**
   * The mask for a byte.
   */
  private static final int BYTE_MASK = 0xff;

  /**
   * The value added to the byte before conversion to String.
   */
  private static final int ADDITION = 0x100;

  /**
   * The hexadecimal radix.
   */
  private static final int RADIX = 16;

  /**
   * The index specifying the starting index for the substring method.
   */
  private static final int BEGIN_INDEX = 1;

  /**
   * The name of the hashing algorithm.
   */
  private static final String SHA_512_ALGORITHM_NAME = "SHA-512";

  /**
   * The properties class that is used to return externally stored secret salt.
   */
  private final SecurityConfigProperties securityConfigProperties;

  /**
   * Encrypts a given email address by salting it with a fixed salt and hashing
   * it afterwards.
   *
   * @param email The email address that needs to be salted and hashed
   * @return The salted and hashed email address
   */
  public String saltAndHashEmail(final String email) {
    String generatedEmail = null;
    try {
      MessageDigest md = MessageDigest.getInstance(SHA_512_ALGORITHM_NAME);
      md.update(getSalt());
      byte[] bytes = md.digest(email.getBytes(StandardCharsets.UTF_8));
      StringBuilder sb = new StringBuilder();
      for (byte b : bytes) {
        sb.append(Integer
            .toString((b & BYTE_MASK) + ADDITION, RADIX)
            .substring(BEGIN_INDEX));
      }
      generatedEmail = sb.toString();
    } catch (NoSuchAlgorithmException e) {
      e.printStackTrace();
    }
    return generatedEmail;
  }

  /**
   * Returns the salt for encrypting email addresses in the form of
   * base64-decoded bytes of a locally stored secret signing key.
   *
   * @return {@code byte[]} containing the salt
   */
  private byte[] getSalt() {
    return Decoders.BASE64.decode(securityConfigProperties.emailSigningKey());
  }

}
