package org.psesquared.server.config;

import io.jsonwebtoken.ExpiredJwtException;
import jakarta.servlet.FilterChain;
import jakarta.servlet.ServletException;
import jakarta.servlet.http.Cookie;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import java.io.IOException;
import lombok.RequiredArgsConstructor;
import org.springframework.lang.NonNull;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.security.core.userdetails.UserDetailsService;
import org.springframework.security.web.authentication.WebAuthenticationDetailsSource;
import org.springframework.stereotype.Component;
import org.springframework.web.filter.OncePerRequestFilter;
import org.springframework.web.util.WebUtils;

/**
 * This filter class handles authentication via JWT.
 * <br>
 * Its method
 * {@link
 * #doFilterInternal(HttpServletRequest, HttpServletResponse, FilterChain)}
 * is invoked before the mapping of the request to the Controller happens.
 */
@Component
@RequiredArgsConstructor
public class JwtAuthenticationFilter extends OncePerRequestFilter {

  /**
   * The URL for the unsecured register-API-endpoint.
   */
  private static final String REGISTER_URL
      = "/api/2/auth/register.json";

  /**
   * The URL for the unsecured forgotPassword-API-endpoint.
   */
  private static final String FORGOT_URL
      = "/api/2/auth/{email}/forgot.json";

  /**
   * The URL for the unsecured verify-API-endpoint.
   */
  private static final String VERIFY_URL
      = "/api/2/auth/{username}/verify.json";

  /**
   * The URL for the unsecured resetPassword-API-endpoint.
   */
  private static final String RESET_PASSWORD_URL
      = "/api/2/auth/{username}/resetpassword.json";

  /**
   * The name of the cookie used for JWT authentication.
   */
  private static final String COOKIE_NAME = "sessionid";

  /**
   * The service class used for managing JWTs.
   */
  private final JwtService jwtService;

  /**
   * The service class used for retrieving users from the database.
   */
  private final UserDetailsService userDetailsService;

  /**
   * The filter method does nothing for the specified unsecured URLs
   * and otherwise calls
   * {@link #authenticateIfValid(Cookie, HttpServletRequest)}.
   *
   * @param request The {@link HttpServletRequest}
   * @param response The {@link HttpServletResponse}
   * @param filterChain The {@link FilterChain}
   * @throws ServletException If error occurs when processing request
   * @throws IOException If I/O error occurs
   */
  @Override
  protected void doFilterInternal(@NonNull final HttpServletRequest request,
                                  @NonNull final HttpServletResponse response,
                                  @NonNull final FilterChain filterChain)
      throws ServletException, IOException {

    final Cookie cookie = WebUtils.getCookie(request, COOKIE_NAME);
    final String url = request.getRequestURI();

    if (url.equals(REGISTER_URL) || url.equals(FORGOT_URL)
        || url.equals(VERIFY_URL) || url.equals(RESET_PASSWORD_URL)
        || cookie == null) {
      filterChain.doFilter(request, response);
      return;
    }

    authenticateIfValid(cookie, request);
    filterChain.doFilter(request, response);
  }

  /**
   * Authenticates the {@link org.psesquared.server.model.User} associated with
   * the JWT from the cookie if it is valid.
   *
   * @param cookie The cookie containing the JWT
   * @param request The {@link HttpServletRequest} for creating a new
   *                authentication details instance
   */
  private void authenticateIfValid(final Cookie cookie,
                                   final HttpServletRequest request) {
    final String jwt = cookie.getValue();
    final String usernameFromToken;

    try {
      usernameFromToken = jwtService.extractAuthUsername(jwt);
    } catch (ExpiredJwtException e) {
      return;
    }

    if (usernameFromToken != null
        && SecurityContextHolder
        .getContext()
        .getAuthentication() == null) {
      UserDetails userDetails
          = userDetailsService.loadUserByUsername(usernameFromToken);
      if (jwtService.isAuthTokenValid(jwt, userDetails)) {
        UsernamePasswordAuthenticationToken authToken
            = new UsernamePasswordAuthenticationToken(
                userDetails,
                null,
                userDetails.getAuthorities()
            );
        authToken.setDetails(
            new WebAuthenticationDetailsSource().buildDetails(request)
        );
        SecurityContextHolder.getContext().setAuthentication(authToken);
      }
    }
  }

}
