package org.psesquared.server.subscriptions.api.controller;

import com.fasterxml.jackson.annotation.JsonProperty;
import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.List;
import lombok.NonNull;

/**
 * SubscriptionDeltas contain all changes that were made to the subscriptions
 * of a user (added / removed podcasts) at a certain time.
 */
public class SubscriptionDelta {

  /**
   * The list of recently subscribed podcasts.
   */
  @JsonProperty(required = true)
  @NonNull
  private final List<String> add;

  /**
   * The list of recently unsubscribed podcasts.
   */
  @JsonProperty(required = true)
  @NonNull
  private final List<String> remove;

  /**
   * The timestamp of the delta.
   */
  @JsonProperty(access = JsonProperty.Access.READ_ONLY)
  private final long timestamp;

  /**
   * Instantiates a new SubscriptionDelta with a current timestamp.
   *
   * @param addedPodcastUrls    List of Strings containing the RSS-Feed URLs of
   *                            all added podcasts
   * @param removedPodcastUrls List of Strings containing the RSS-Feed URLs of
   *                           all removed podcasts
   */
  public SubscriptionDelta(
      @org.springframework.lang.NonNull final List<String> addedPodcastUrls,
      @org.springframework.lang.NonNull final List<String> removedPodcastUrls) {
    this.add = addedPodcastUrls;
    this.remove = removedPodcastUrls;
    this.timestamp = LocalDateTime.now().toEpochSecond(ZoneOffset.UTC);
  }

  /**
   * Returns the list of RSS-Feed URLs of all added podcasts.
   *
   * @return RSS-Feed URLs of all added podcasts
   */
  @org.springframework.lang.NonNull
  public List<String> getAdd() {
    return add;
  }

  /**
   * Returns the list of RSS-Feed URLs of all removed podcasts.
   *
   * @return RSS-Feed URLs of all removed podcasts
   */
  @org.springframework.lang.NonNull
  public List<String> getRemove() {
    return remove;
  }

  /**
   * Returns the timestamp of when this Subscription Delta was uploaded.
   *
   * @return The timestamp of when this Subscription Delta was uploaded
   */
  public long getTimestamp() {
    return timestamp;
  }

}
